use gstreamer::glib::{self, ParamSpec, ParamSpecBoolean, ParamSpecString, Value};
use gstreamer::prelude::*;
use gstreamer::subclass::prelude::*;
use gstreamer::subclass::ElementMetadata;
use gstreamer::{
    Buffer, Caps, CapsIntersectMode, DebugCategory, DebugColorFlags, ErrorMessage, FlowError,
    FlowSuccess, LoggableError, PadDirection, PadPresence, PadTemplate, ResourceError,
};

use gstreamer_base::subclass::prelude::*;

use gstreamer_video::prelude::*;
use gstreamer_video::subclass::prelude::*;
use gstreamer_video::video_frame::Readable;
use gstreamer_video::{VideoCapsBuilder, VideoFormat, VideoFrame, VideoInfo, VideoSink};

use std::cmp;
use std::net::{Ipv6Addr, SocketAddr, UdpSocket};
use std::sync::{LazyLock, Mutex};

const HEIGHT: u32 = 196;

const WIDTH: u32 = 346;

static CAT: LazyLock<DebugCategory> = LazyLock::new(|| {
    DebugCategory::new(
        "pingxelflutsink",
        DebugColorFlags::empty(),
        Some("Pingxelflut sink"),
    )
});

#[derive(Default)]
struct Settings {
    bind_address: Option<SocketAddr>,
    diff_only: bool,
    prefix: Option<Ipv6Addr>,
}

#[derive(Default)]
enum State {
    #[default]
    Stopped,
    Started {
        socket: UdpSocket,
    },
}

#[derive(Default)]
pub struct PingxelflutSink {
    caps: Mutex<Option<Caps>>,
    info: Mutex<Option<VideoInfo>>,
    prev_frame: Mutex<Option<VideoFrame<Readable>>>,
    settings: Mutex<Settings>,
    state: Mutex<State>,
}

#[glib::object_subclass]
impl ObjectSubclass for PingxelflutSink {
    const NAME: &'static str = "GstPingxelflutSink";

    type Type = super::PingxelflutSink;
    type ParentType = VideoSink;
}

impl ObjectImpl for PingxelflutSink {
    fn properties() -> &'static [ParamSpec] {
        static PROPERTIES: LazyLock<Vec<ParamSpec>> = LazyLock::new(|| {
            vec![
                ParamSpecString::builder("bind-address")
                    .nick("Bind address")
                    .blurb("Local address to bind to")
                    .build(),
                ParamSpecBoolean::builder("diff-only")
                    .nick("Diff only")
                    .blurb("Only send pixels different from previous frame")
                    .build(),
                ParamSpecString::builder("prefix")
                    .nick("Destination prefix")
                    .blurb("Prefix of pingxelflut server")
                    .build(),
            ]
        });

        PROPERTIES.as_ref()
    }

    fn set_property(&self, _id: usize, value: &Value, pspec: &ParamSpec) {
        let mut settings = self.settings.lock().unwrap();

        match pspec.name() {
            "bind-address" => {
                settings.bind_address = value
                    .get::<Option<String>>()
                    .unwrap()
                    .map(|s| s.parse().ok())
                    .unwrap_or(None);
            }
            "diff-only" => {
                settings.diff_only = value.get::<bool>().unwrap();
            }
            "prefix" => {
                settings.prefix = value
                    .get::<Option<String>>()
                    .unwrap()
                    .map(|s| s.parse().ok())
                    .unwrap_or(None);
            }
            _ => unimplemented!(),
        }
    }

    fn property(&self, _id: usize, pspec: &ParamSpec) -> Value {
        let settings = self.settings.lock().unwrap();

        match pspec.name() {
            "bind-address" => settings
                .bind_address
                .as_ref()
                .map(|addr| addr.to_string())
                .to_value(),
            "diff-only" => settings.diff_only.to_value(),
            "prefix" => settings
                .prefix
                .as_ref()
                .map(|addr| addr.to_string())
                .to_value(),
            _ => unimplemented!(),
        }
    }
}

impl GstObjectImpl for PingxelflutSink {}

impl ElementImpl for PingxelflutSink {
    fn metadata() -> Option<&'static ElementMetadata> {
        static ELEMENT_METADATA: LazyLock<ElementMetadata> = LazyLock::new(|| {
            ElementMetadata::new(
                "Pingxelflut sink",
                "Sink/Video",
                "Send stream to a pingxelflut screen",
                "Luca <Luca@hackerspace-bamberg.de>",
            )
        });

        Some(&*ELEMENT_METADATA)
    }

    fn pad_templates() -> &'static [PadTemplate] {
        static PAD_TEMPLATES: LazyLock<Vec<PadTemplate>> = LazyLock::new(|| {
            vec![PadTemplate::new(
                "sink",
                PadDirection::Sink,
                PadPresence::Always,
                &VideoCapsBuilder::new().format(VideoFormat::Rgb).build(),
            )
            .unwrap()]
        });

        PAD_TEMPLATES.as_ref()
    }
}

impl BaseSinkImpl for PingxelflutSink {
    fn start(&self) -> Result<(), ErrorMessage> {
        let mut state = self.state.lock().unwrap();
        if let State::Started { .. } = *state {
            unreachable!("PingxelflutSink already started");
        }

        let settings = self.settings.lock().unwrap();
        let bind_address = settings.bind_address.as_ref().ok_or_else(|| {
            gstreamer::error_msg!(ResourceError::Settings, ["Bind address is not defined"])
        })?;

        if let None = settings.prefix {
            return Err(gstreamer::error_msg!(
                ResourceError::Settings,
                ["Prefix is not defined"]
            ));
        }

        let socket = UdpSocket::bind(bind_address).map_err(|_| {
            gstreamer::error_msg!(
                ResourceError::OpenWrite,
                ["Could not bind UDP socket to address {}", bind_address]
            )
        })?;

        *state = State::Started { socket };

        Ok(())
    }

    fn stop(&self) -> Result<(), ErrorMessage> {
        let mut state = self.state.lock().unwrap();
        if let State::Stopped = *state {
            return Err(gstreamer::error_msg!(
                ResourceError::Settings,
                ["PingxelflutSink not started"]
            ));
        }

        *state = State::Stopped;

        Ok(())
    }

    fn caps(&self, filter: Option<&Caps>) -> Option<Caps> {
        let mut caps = self
            .caps
            .lock()
            .unwrap()
            .clone()
            .unwrap_or_else(|| Self::pad_templates()[0].caps().clone());

        if let Some(filter_caps) = filter {
            caps = filter_caps.intersect_with_mode(&caps, CapsIntersectMode::First);
        }

        Some(caps)
    }

    fn set_caps(&self, caps: &Caps) -> Result<(), LoggableError> {
        let video_info = VideoInfo::from_caps(caps)
            .map_err(|_| gstreamer::loggable_error!(CAT, "Invalid caps"))?;

        *self.info.lock().unwrap() = Some(video_info);

        Ok(())
    }
}

impl VideoSinkImpl for PingxelflutSink {
    fn show_frame(&self, buffer: &Buffer) -> Result<FlowSuccess, FlowError> {
        let info = self.info.lock().unwrap();
        let info = info.as_ref().ok_or_else(|| FlowError::NotNegotiated)?;

        let frame =
            VideoFrame::from_buffer_readable(buffer.clone(), info).map_err(|_| FlowError::Error)?;
        let data = frame.plane_data(0).unwrap();

        let prev_frame = self.prev_frame.lock().unwrap().take();

        let settings = self.settings.lock().unwrap();
        let prefix = settings.prefix.as_ref().unwrap();

        if let State::Started { ref socket } = *self.state.lock().unwrap() {
            for y in 0..cmp::min(frame.height(), HEIGHT) {
                for x in 0..cmp::min(frame.width(), WIDTH) {
                    let i =
                        (y * frame.plane_stride()[0] as u32 + x * frame.n_components()) as usize;
                    let (r, g, b) = (data[i] as u16, data[i + 1] as u16, data[i + 2] as u16);

                    if settings.diff_only
                        && prev_frame.as_ref().map_or(false, |f| {
                            f.plane_data(0).is_ok_and(|d| {
                                d[i] as u16 == r && d[i + 1] as u16 == g && d[i + 2] as u16 == b
                            })
                        })
                    {
                        continue;
                    }

                    if let Err(err) = socket.send_to(
                        &[],
                        (
                            prefix
                                | Ipv6Addr::new(
                                    0,
                                    0,
                                    0,
                                    0,
                                    x as u16,
                                    y as u16,
                                    (r << 8) | g,
                                    b << 8,
                                ),
                            1337,
                        ),
                    ) {
                        gstreamer::warning!(CAT, imp = self, "{}", err);
                    }
                }
            }
        }

        self.prev_frame.lock().unwrap().replace(frame);

        Ok(FlowSuccess::Ok)
    }
}
